'''Use actual data.'''

import numpy as np
import matplotlib.pyplot as plt
from scipy.io import loadmat
from t1est import t1est

if __name__ == '__main__':

    fft = lambda x0, axes: np.fft.ifftshift(np.fft.fft2(
        np.fft.fftshift(x0, axes=axes),
        axes=axes), axes=axes)
    ifft = lambda x0, axes: np.fft.fftshift(np.fft.ifft2(
        np.fft.ifftshift(x0, axes=axes),
        axes=axes), axes=axes)

    path = (
        '/home/nicholas/Documents/rawdata/'
        'P043018-20190710T193153Z-001/P043018/')
    filename = 'FID12461.mat'
    d = loadmat(path + filename)['kspace']
    d = ifft(d, axes=(0, 1))
    time_frames = [0, 5, 1, 6, 2, 7, 3, 4]
    d = d[..., time_frames, :]
    sx4 = int(d.shape[0]/4)
    d = d[sx4:-sx4, ...]
    sx, sy, st, sc = d.shape[:]

    # Get magnitude data
    mag = True
    d = np.sqrt(np.sum(np.abs(d)**2, axis=-1))
    d = 100*d/np.max(d.flatten())

    # # Use complex data
    # mag = False
    # phi = np.zeros((sx, sy, st))
    # for tt in range(st):
    #     phi[..., tt] = virtcoilphase(d[..., tt])
    # d = np.sqrt(np.sum(np.abs(d)**2, axis=-1))
    # d = d*phi
    # plt.imshow(np.abs(d[..., 0]))
    # plt.show()

    TIs = np.array([
        117., 257., 1172., 1282., 2172., 2325., 3174., 4189.])
    TIs *= 1e-3

    mask = np.zeros((sx, sy), dtype=bool)
    sx4, sy4 = int(sx/3), int(sy/3)
    mask[sx4:-sx4, sy4:-sy4] = True

    T1map = t1est(
        d, TIs, time_axis=-1, mask=mask, method='lm',
        T1_bnds=(0, 3), chunksize=10, molli=True, mag=mag)

    plt_args = {'vmin':0, 'vmax':3}
    plt.imshow(T1map, **plt_args)
    plt.show()



'''Helper func.'''

import numpy as np
from scipy.io import loadmat

def loader(idx):
    '''Load data, inversion times, and truth from MAT files.'''

    basename = 'data/Pre_contrast_MOLLI_T1_testing'
    filenames = [
        'P072919_T1Map_LongT1_2CH_MOCO_19.mat',
        'P072919_T1Map_LongT1_4CH_MOCO_16.mat',
        'P072919_T1Map_LongT1_SA_Apex_MOCO_13.mat',
        'P072919_T1Map_LongT1_SA_Base_MOCO_7.mat',
        'P072919_T1Map_LongT1_SA_Mid_MOCO_10.mat',
        'P082119_T1Map_LongT1_4CH_MOCO_16.mat',
        'P082119_T1Map_LongT1_SA_Apex_MOCO_13.mat',
        'P082119_T1Map_LongT1_SA_Apex_MOCO_45.mat',
        'P082119_T1Map_LongT1_SA_Base_MOCO_7.mat',
        'P082119_T1Map_LongT1_SA_Mid_MOCO_10.mat',
        'P082119_T1Map_LongT1_SA_Mid_MOCO_42.mat',
        'P082819_T1Map_LongT1_4CH_MOCO_20.mat',
        'P082819_T1Map_LongT1_SA_Apex_MOCO_17.mat',
        'P082819_T1Map_LongT1_SA_Base_MOCO_8.mat',
        'P082819_T1Map_LongT1_SA_Mid_MOCO_11.mat',
        'P082819_T1Map_LongT1_SA_Mid_MOCO_14.mat',
    ]

    data = loadmat(basename + '/' + filenames[idx])
    TIs = data['inv_times'].squeeze()

    # remove all other keys except the one we want
    for key in [
            '__version__', '__header__', '__globals__', 'inv_times']:
        del data[key]

    # The only remaining key should be the data
    key = list(data.keys())[0]

    # First eight are the images corresponding to inversion times
    ims = data[key][..., :8].astype(np.double)/1000

    # Last is the reference T1
    truth = data[key][..., -1].astype(np.double)/1000

    # Sort TIs and ims to be chronological
    ind = np.argsort(TIs)
    TIs = np.sort(TIs)
    ims = ims[..., ind]

    # We want TIs in seconds
    TIs = TIs*1e-3

    return(TIs, ims, truth, filenames[idx][:-4])



'''Run datasets from Ganesh and compare to Siemens' truth.'''

from time import time
from os.path import isfile

import numpy as np
import matplotlib.pyplot as plt
from scipy.io import savemat
from skimage.measure import compare_nrmse, compare_ssim
from skimage.filters import threshold_li
from skimage.morphology import convex_hull_image

from t1est import t1est
from .loader import loader

def run(
        TIs, ims, truth, mask, method, chunksize=50, filename=None,
        show=True, force=False):
    '''Run a dataset.'''

    if not isfile(filename + '.npz') or force:
        T1map = t1est(
            ims, TIs, time_axis=-1, mask=mask, method=method,
            T1_bnds=(0, 5), chunksize=chunksize, molli=True, mag=True)
        T1map = np.nan_to_num(T1map)

        # Assume that anything estimated above 5s is wrong
        T1map[np.abs(T1map) > 5] = 0

        # Assume that anything estimated below 0s is wrong
        T1map[T1map < 0] = 0

        np.savez_compressed(
            filename + '.npz',
            mask=mask,
            method=method,
            T1map=T1map)

    else:
        print('Result exists!')
        data = np.load(filename + '.npz')
        # savemat(filename + '.mat', mdict=data)
        T1map = data['T1map']


    # Some plots
    plt.figure(figsize=(10, 3))
    plt_args = {'vmin':0, 'vmax':4}
    nx, ny = 1, 3
    ax = plt.subplot(nx, ny, 1)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    plt.imshow(truth, **plt_args)
    plt.title('Truth')

    ax = plt.subplot(nx, ny, 2)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    plt.imshow(T1map, **plt_args)
    plt.title('T1 est (%s)' % method)

    ax = plt.subplot(nx, ny, 3)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    plt.imshow(truth - T1map, **plt_args)
    nrmse = compare_nrmse(truth, T1map)
    ssim = compare_ssim(truth, T1map)
    print('NRMSE: %g' % nrmse)
    print(' SSIM: %g' % ssim)
    plt.title('NRMSE: %g, SSIM: %g' % (nrmse, ssim))

    if filename:
        plt.savefig('%s.png' % filename)
    if show:
        plt.show()

def find_colorbar(idx):
    '''Give back location of colorbar'''

    # The colorbar appears to be in one of two positions:
    pos0 = (67, 112, 5, 16)
    pos1 = (76, 128, 4, 13)
    return [
        pos0, pos0, pos1, pos0, pos0, pos0, pos1, pos1, pos1, pos1,
        pos1, pos0, pos1, pos1, pos1, pos1][idx]

if __name__ == '__main__':

    # num_pts = 8
    num_pts = 5

    for idx in range(16):

        print('STARTING %d' % idx)
        TIs, ims, truth, name = loader(idx)

        # Pop images off of the front until we have the ones we want
        if ims.shape[-1] > num_pts:
            nt = ims.shape[-1]
            # step = int(np.ceil(nt/num_pts))
            # ind = [ii for ii in range(0, nt, step)]
            # if len(ind) == num_pts-1:
            #     ind.insert(int(len(ind)/2), int(nt/2)-1)
            # print(ind)
            # ims = ims[..., ind]
            # TIs = TIs[ind]
            # print(ims.shape, TIs.shape)

            # Use first 5 time points
            ind = [ii for ii in range(num_pts)]
            ims = ims[..., ind]
            TIs = TIs[ind]
            print(ims.shape, TIs.shape)


        t0 = time()
        sos = np.sqrt(np.sum(np.abs(ims)**2, axis=-1))
        thresh = threshold_li(sos)
        mask = convex_hull_image(sos > thresh)

        # Mask should always include Siemens' colorbar
        x, xx, y, yy = find_colorbar(idx)
        mask[x:xx, y:yy] = False

        truth = truth*mask
        print('Took %g seconds to get mask' % (time() - t0))

        run(
            TIs, ims, truth, mask, method='lm', chunksize=50,
            filename='data/results/lm_first_t%d_%s' % (num_pts, name),
            show=False, force=False)

        # run(
        #     TIs, ims, truth, mask, method='trf', chunksize=20,
        #     filename='data/results/trf_%s' % name, show=False)



'''Show basic usage of t1est.'''

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from phantominator import shepp_logan
from t1est import t1est

if __name__ == '__main__':

    # Make a simple Shepp-Logan MR phantom
    N = 64
    M0 = shepp_logan(N)
    T1 = 1 + M0*2

    # Naive inversion recovery simulation for this demo
    nt = 11
    TIs = np.linspace(.01, 3*np.max(T1.flatten()), nt)
    ph = M0[..., None]*(
        1 - 2*np.exp(-TIs[None, None, :]/T1[..., None]))
    mask = np.abs(ph[..., -1]) > 1e-8

    # Get a T1 map!
    T1map = t1est(
        ph, TIs, time_axis=-1, mask=mask, method='lm',
        T1_bnds=(np.min(T1.flatten()), np.max(T1.flatten())),
        chunksize=10, molli=False, mag=False)

    # Check it out
    nx, ny = 1, 3
    cbar_opts = {'size':"5%", 'pad':0.05}
    plt_opts = {'vmin': 0, 'vmax': 3}

    ax = plt.subplot(nx, ny, 1)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    im = plt.imshow(T1*mask, **plt_opts)
    plt.title('Target T1 map')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', **cbar_opts)
    plt.colorbar(im, cax=cax)

    ax = plt.subplot(nx, ny, 2)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    im = plt.imshow(T1map, **plt_opts)
    plt.title('T1 estimates')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)


    ax = plt.subplot(nx, ny, 3)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    im = plt.imshow(np.abs((T1map - T1)*mask))
    plt.title('Residual')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)
    plt.show()



'''Setup.py'''

from distutils.core import setup
from setuptools import find_packages

setup(
    name='t1est',
    version='0.1.2',
    author='Nicholas McKibben',
    author_email='nicholas.bgp@gmail.com',
    packages=find_packages(),
    scripts=[],
    url='https://github.com/mckib2/t1est',
    license='GPLv3',
    description='Basic T1 fitting',
    long_description=open('README.rst').read(),
    install_requires=[
        "numpy>=1.17.2",
        "matplotlib>=3.1.1",
        "phantominator>=0.4.3",
        "tqdm>=4.36.1",
    ],
    python_requires='>=3.6',
)



'''Simple T1 estimation.'''

from functools import partial
from multiprocessing import Pool

import numpy as np
from scipy.optimize import least_squares
from tqdm import tqdm

def t1est(
        x, TIs, time_axis=-1, mask=None, method='lm', T1_bnds=None,
        chunksize=10, molli=False, mag=True):
    '''Make a T1 map.

    Parameters
    ----------
    x : array_like
        Image space data.
    TIs : array_like
        Inversion times.
    time_axis : int, optional
        Dimension of x that holds the time points.
    mask : array_like
        In-plane mask.
    method : {'trf', 'lm'}, optional
        Trust region reflective or Lev-Mar for fitting nonlinear
        least squares. Lev-Mar is faster, trust region is more
        accurate and can use bounds.
    T1_bnds : tuple or None, optional
        Upper and lower bound for T1 to use when fitting.  Only for
        method='trf'.
    chunksize : int, optional
        Multiprocessing chunksize.
    molli : bool, optional
        Do MOLLI correction.  T1_bnds will be considered as bounds on
        apparent T1.
    mag : bool, optional
        Magnitude data or complex data.

    Returns
    -------
    T1map : array_like
        T1 estimates the same shape as mask.
    '''

    # Move time points to the back of the bus
    x = np.moveaxis(x, time_axis, -1)
    TIs = np.array(TIs)
    nt = x.shape[-1]
    sh = x.shape[:-1]

    # If we don't have a mask, just do everything
    if mask is None:
        mask = np.ones(sh, dtype=bool)

    # Sanity checks
    assert nt == TIs.size, (
        'Number of inversion times does not match the number of '
        'time frames in x!')
    assert mask.shape == sh, 'mask should match x!'

    # Get bounds and initial guess for least squares solver
    if mag:
        x0 = np.array([1, 2, 1]) # better A, B initial estimates
    else:
        x0 = np.ones(5)
    if method == 'trf':
        if T1_bnds is None:
            T1_bnds = (0, np.inf)
        if mag:
            bnds = (
                (-np.inf, -np.inf, T1_bnds[0]),
                (np.inf, np.inf, T1_bnds[1]))
        else:
            bnds = (
                (-np.inf, -np.inf, -np.inf, -np.inf, T1_bnds[0]),
                (np.inf, np.inf, np.inf, np.inf, T1_bnds[1]))
    elif method == 'lm':
        bnds = (-np.inf, np.inf)
    else:
        raise NotImplementedError()

    # Parallelize this thing!
    ntot = np.prod(sh)
    T1map = np.zeros(ntot)
    mask = mask.flatten()
    x = np.reshape(x, (-1, nt))
    idx = np.argwhere(mask).squeeze()
    T1map = T1map.flatten()
    fitp = partial(
        _fitwrap, x=x, TIs=TIs, method=method, bnds=bnds,
        molli=molli, mag=mag, x0=x0)
    with Pool() as pool:
        res = list(tqdm(
            pool.imap(fitp, idx, chunksize),
            total=idx.size, leave=False))
    T1map[idx] = np.array(res)
    return np.reshape(T1map, sh)

def _fitwrap(idx, x, TIs, method, bnds, molli, mag, x0):
    '''Picklable wrapper for _fit.'''
    return _fit(x[idx, :], TIs, method, bnds, molli, mag, x0)

def _model(Ar, Ai, Br, Bi, T1, TI):
    '''T1 fitting model with complex coefficients.'''
    return (Ar + 1j*Ai) - (Br + 1j*Bi)*np.exp(-TI/T1)

def _magmodel(A, B, T1, TI):
    '''T1 fitting model with real coefficients.'''
    return np.abs(A - B*np.exp(-TI/T1))

def _obj(x, t, y):
    '''Function for least squares fitting'''
    Ar, Ai, Br, Bi, T1 = x[:]
    return np.abs(_model(Ar, Ai, Br, Bi, T1, t) - y)

def _magobj(x, t, y):
    '''Function for least squares fitting assuming magnitude data.'''
    A, B, T1 = x[:]
    return np.abs(_magmodel(A, B, T1, t) - y)

def _fit(y, t, method, bnds, molli, mag, x0):
    '''Do T1 fitting'''

    obj = _obj
    if mag:
        obj = _magobj

    old_settings = np.seterr(over='ignore')
    res_lsq = least_squares(
        obj, x0, args=(t, y), method=method, bounds=bnds)
    np.seterr(**old_settings)

    if mag:
        A, B, T1 = res_lsq['x'][:]
    else:
        Ar, Ai, Br, Bi, T1 = res_lsq['x'][:]

    if molli:
        if mag:
            return T1*(B/A - 1)
        return T1*(np.abs(Br + 1j*Bi)/np.abs(Ar + 1j*Ai) - 1)
    return T1

if __name__ == '__main__':
    pass
